#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <jasper/jasper.h>
#include "grb2.h"
#include "wgrib2.h"
#include "fnlist.h"

#include "grib2.h"
int enc_jpeg2000(unsigned char *cin, g2int width,g2int height,g2int nbits, g2int ltype, 
   g2int ratio, g2int retry, char *outjpc, g2int jpclen);

/*
 * Grib_out
 *
 * routines to encode data into grib2
 *
 * 12/2007 Public Domain by Wesley Ebisuzaki
 *
 */

unsigned char *mk_bms(float *data, unsigned int *ndata);
int mk_sec5_7(float *data, unsigned int n, 
	unsigned char **sec5, unsigned char **sec7);

static int packing_nx(float *data, int nx, int ny, unsigned int ndata);


extern int decode, nx, ny, scan;
extern int flush_mode;
extern enum output_order_type output_order;
extern int use_scale, dec_scale, bin_scale;
extern int save_translation;


static int max_bits = 16;
static int wanted_bits = 12;
enum {jpeg,ieee,simple} grib_type = simple;

/*
 * HEADER:100:set_grib_type:misc:1:set grib type = jpeg, simple, ieee
 */

int f_set_grib_type(ARG1) {

    if (strcmp(arg1,"jpeg") == 0) grib_type = jpeg;
    else if (strcmp(arg1,"ieee") == 0) grib_type = ieee;
    else if (strcmp(arg1,"simple") == 0) grib_type = simple;
    else fatal_error("set_grib_type: bad type %s", arg1);
    return 0;
}
 
/*
 * HEADER:100:grib_out:output:1:writes decoded/modified data in grib-2 format to file X
 */

int f_grib_out(ARG1) {

    float *data_tmp;

    if (mode == -1) {
        save_translation = decode = 1;
        if ((*local = (void *) ffopen(arg1, "wb")) == NULL) 
		fatal_error("Could not open %s", arg1);
    }
    else if (mode >= 0) {
	if ((data_tmp = (float *) malloc(ndata * sizeof(float))) == NULL)
	 	fatal_error("memory allocation - data_tmp","");
        undo_output_order(data, data_tmp, ndata);

        if (grib_type == simple) grib_out(sec, data_tmp, ndata, (FILE *)*local);
        else if (grib_type == jpeg) jpeg_grib_out(sec, data_tmp, ndata, nx, 
                     ny, use_scale, dec_scale, bin_scale, (FILE *)*local);
        else if (grib_type == ieee) ieee_grib_out(sec, data_tmp, ndata, 
		(FILE *)*local);
        if (flush_mode) fflush((FILE *) *local);
	free(data_tmp);
    }
    return 0;
}

/*
 * write grib-2 ieee file
 */

int ieee_grib_out(unsigned char **sec, float *data, unsigned int ndata, FILE *out) {

    unsigned int n_defined, i;
    int j;
    unsigned char *p, *sec0, *sec1, *sec2, *sec3, *sec4, *sec5, *sec6, *sec7;
#ifdef IEEE_BITMAP
    float *data_tmp;
#endif

    /* required passed sections */
    sec0 = sec[0];
    sec1 = sec[1];
    sec2 = sec[2];
    sec3 = sec[3];
    sec4 = sec[4];

    /* change scan mode */
    // flag = flag_table_3_4(sec);
    // set_order(sec, output_order);

    /* make a new section 6 */

#ifdef IEEE_BITMAP
    data_tmp = (float *) malloc(ndata * sizeof(float));
    for (i = 0; i < ndata; i++) {
	data_tmp[i[ = data[i];
    }
    n_defined = ndata;
    sec6 = mk_bms(data_tmp, &n_defined);			// make bitmap section
    if (sec6 == NULL) fatal_error("grib_out ieee memory allocation sec6","");
#else
    n_defined = ndata;
    sec6 = (unsigned char *) malloc(6);
    if (sec6 == NULL) fatal_error("grib_out ieee memory allocation sec6","");
    uint_char(6 * sizeof (unsigned char), sec6);
    sec6[4] = 6;			// section 5
    sec6[5] = 255;			// no bitmap
#endif
    
    /* data representation section */

    sec5 = (unsigned char *) malloc(12 * sizeof(unsigned char));
    if (sec5 == NULL) fatal_error("grib_out ieee memory allocation sec5","");
    uint_char(12 * sizeof (unsigned char), sec5);
    sec5[4] = 5;			// section 5
    uint_char(ndata, sec5+5);		// number of points
    uint2_char(4,sec5+9);		// data template 4
    sec5[11] = 1;			// precision: ieee 32-bit


    /* data section */
    sec7 = (unsigned char *) malloc(5 + n_defined * 4);
    if (sec7 == NULL) fatal_error("grib_out ieee memory allocation sec7","");
    uint_char(5+n_defined*4, sec7);
    sec7[4] = 7;
    p = sec7 + 5;
    j = 0;
    for (i = 0; i < n_defined; i++) {
#ifdef IEEE_BITMAP
	flt2ieee_nan(data_tmp[i], p);
#else
	flt2ieee_nan(data[i], p);
#endif
	p += 4;
    }
#ifdef IEEE_BITMAP
    free(data_tmp);
#endif

    j = wrt_sec(sec0, sec1, sec2, sec3, sec4, sec5, sec6, sec7, out);

    free(sec5);
    free(sec6);
    free(sec7);

    /* set scan mode to original order */
    // set_flag_table_3_4(sec, flag);
    
    return j;

}
/*
 * write a grib-2 file 
 *
 * sec0..sec4 predefined sections 0 to 4
 * data[] = values to encode into grib
 * ndata = size of data
 * out = output file
 *
 * currently this program writes in simple packing 12 bits per data point
 */

int grib_out(unsigned char **sec, float *data, unsigned int ndata, FILE *out) {

    unsigned int n_defined;
    int i;
    unsigned char *sec0, *sec1, *sec2 , *sec3, *sec4, *sec5, *sec6, *sec7;

    /* required passed sections */
    sec0 = sec[0];
    sec1 = sec[1];
    sec2 = sec[2];
    sec3 = sec[3];
    sec4 = sec[4];

    /* change scan mode */
    // flag = flag_table_3_4(sec);
    // set_order(sec, output_order);

    /* make a sections 5-7 */
    n_defined = ndata;
    sec6 = mk_bms(data, &n_defined);			// make bitmap section
    mk_sec5_7(data, n_defined, &sec5, &sec7);		// make sec 5 and 7

    i = wrt_sec(sec0, sec1, sec2, sec3, sec4, sec5, sec6, sec7, out);

    free(sec5);
    free(sec6);
    free(sec7);

    /* set scan mode to original order */
    // set_flag_table_3_4(sec, flag);

    return i;
}

/*
 *  writes out jpeg2000 compressed grib message
 */

int jpeg_grib_out(unsigned char **sec, float *data, unsigned int ndata, 
  int nx, int ny, int use_scale, int dec_scale, int bin_scale, FILE *out) {

    unsigned int n_defined, j;
    int iy, ix, jpclen;
    unsigned char *cdata, *p;
    unsigned char *sec0, *sec1, *sec2 , *sec3, *sec4, *sec5, *sec6, *sec7;
    double fmin, frange, scale, dec_factor;
    float ref, min_val, max_val, ncep_min_val;
    int i, k, nbits, nbytes;

    int ltype, ratio, retry;
    char *outjpc;

    /* required passed sections */
    sec0 = sec[0];
    sec1 = sec[1];
    sec2 = sec[2];
    sec3 = sec[3];
    sec4 = sec[4];

    /* change scan mode */
    // flag = flag_table_3_4(sec);
    // set_order(sec, output_order);

    /* figure out nx and ny */
    ix = packing_nx(data, nx, ny, ndata);

    /* make a sections 5-7 */
    n_defined = ndata;
    sec6 = mk_bms(data, &n_defined);                    // make bitmap section

    if (nx*ny == n_defined) {
        ix = nx;
        iy = ny;
    }
    else {
	iy = n_defined / ix;
	if (ix*iy != n_defined) {
	    ix = n_defined;
 	    iy = 1;
	}
    }


    for (max_val = min_val = data[0], j = 1; j < n_defined; j++) {
        if (min_val > data[j]) min_val = data[j];
        if (max_val < data[j]) max_val = data[j];
    }
    ncep_min_val = min_val;

    if (use_scale == 0) {
        /* ecmwf style */
        fmin = min_val;
        frange = max_val - fmin;
        ref = fmin;
	dec_scale = 0;
        if (frange != 0.0) {
            frexp(frange, &i);
            bin_scale = i - wanted_bits;
            nbits = wanted_bits;
            scale = ldexp(1.0, -bin_scale);
            frange = floor((max_val-fmin)*scale + 0.5);
            frexp(frange, &i);
            if (i != nbits) bin_scale++;
        }
        else {
            bin_scale = nbits = 0;
	    scale = 1;
        }
    }
    else {
	if (dec_scale) {
	    dec_factor = Int_Power(10.0, -dec_scale);
	    min_val *= dec_factor;
	    max_val *= dec_factor;
            for (j = 0; j < n_defined; j++) {
                data[j] *= dec_factor;
            }
	}

        scale = ldexp(1.0, -bin_scale);
	ref = floor(min_val*scale)/scale;
        i = (int) (max_val - ref)*scale + 0.5;
	frange = (double) i;
        frexp(frange, &nbits);
        if (nbits > max_bits) {
	    bin_scale += (nbits - max_bits);
	    nbits = max_bits;
        }
    }
    if (bin_scale) {
        scale = ldexp(1.0, -bin_scale);
        for (j = 0; j < n_defined; j++) {
            data[j] = (data[j] - ref)*scale;
        }
    }
    else {
        for (j = 0; j < n_defined; j++) {
            data[j] = data[j] - ref;
        }
    }

    if (nbits > 0 && n_defined > 0) {

        nbytes = (nbits + 7) / 8;

        /* Pack integers into bytes */

        cdata = malloc(n_defined * nbytes);
        if (cdata == NULL) fatal_error("memory alloc jpeg encoding","");
        p = cdata; 
        if (nbytes == 1) {
            for (j = 0; j < n_defined; j++) {
    	        i = (int) floor(data[j]+0.5);
	        *p++ = i & 255;
	    }
        }
        else if (nbytes == 2) {
            for (j = 0; j < n_defined; j++) {
	        i = (int) floor(data[j]+0.5);
	        p[0] = (i >> 8) & 255;
	        p[1] = i & 255;
	        p += 2;
	    }
        }
        else if (nbytes > 0) {
            for (j = 0; j < n_defined; j++) {
	        i = (int) floor(data[j]+0.5);
	        for (k = 1; k <= nbytes; k++) {
	            p[nbytes - k] = i & 255;
		    i = i >> 8;
	        }
	        p += nbytes;
	    }
        }

//    jas_init();
        ltype = 0;
        ratio = 1;
        retry = 0;

        jpclen = 4*n_defined+200;
        outjpc = malloc(jpclen);

        i = enc_jpeg2000(cdata,ix,iy,nbits,ltype,ratio,retry,outjpc,jpclen);

	free(cdata);
	if (i <= 0) fatal_error_i("enc_jpeg error %d", i);
    }
    else {   // nbits == 0 || n_defined == 0
	i = 0;
	outjpc = malloc(1);
    }

    /* data representation section */

    // fix for buggy NCEP library 1/2009
    // NCEP routines ignore decimal scaling with nbits == 0
    if (nbits == 0) {
	ref = ncep_min_val;
	bin_scale = dec_scale = 0;
    }

    sec5 = (unsigned char *) malloc(23 * sizeof(unsigned char));
    if (sec5 == NULL) fatal_error("grib_out ieee memory allocation sec5","");
    uint_char(23 * sizeof (unsigned char), sec5);
    sec5[4] = 5;                        // section 5
    uint_char(ndata, sec5+5);           // number of points
    uint2_char(40,sec5+9);              // data template 4 - jpeg2000
    flt2ieee(ref,sec5+11);		// reference value
    int2_char(bin_scale,sec5+15);	// binary scaling
    int2_char(-dec_scale,sec5+17);	// decimal scaling
    sec5[19] = nbits;
    sec5[20] = 0;			// 0 - float 1=int
    sec5[21] = 0;			//code 5.40 -lossless
    sec5[22] = 255;			// undefined
    
    sec7 = malloc(i+5);
    if (sec7 == NULL) fatal_error("grib_out jpeg memory allocation sec7","");
    uint_char(i+5, sec7);
    sec7[4] = 7;                        // section 7
    for (k = 0; k < i; k++)
	sec7[5+k] = outjpc[k];

    i = wrt_sec(sec0, sec1, sec2, sec3, sec4, sec5, sec6, sec7, out);

    free(outjpc);
    free(sec5);
    free(sec6);
    free(sec7);

    return i;
}

/*
 * tries to find an appropriate nx for jpeg2000 packing
 */

static int packing_nx(float *data, int nx, int ny, unsigned int ndata) {

    unsigned int i, j;

    for (i = 0; i < ndata; i++) {
	if (data[i] < UNDEFINED_LOW || data[i] > UNDEFINED_HIGH) break;
    }
    if (i == ndata) return 1;
    for (j = i+1; j < ndata; j++) {
	if (data[j] >= UNDEFINED_LOW && data[j] <= UNDEFINED_HIGH) break;
    }
    return j-i;

}

/*
 * based on mk_BMS (from gribw)
 *
 * public domain 12/2007 Wesley Ebisuzaki
 */

unsigned char *mk_bms(float *data, unsigned int *ndata) {

    int bms_size, c, imask;
    unsigned char *bms, *cbits;
    unsigned int nn, i, start;

    nn = *ndata;
    bms_size = 6 + (nn+7) / 8;
    bms = (unsigned char *) malloc(bms_size);

    uint_char(bms_size, bms);		// length of section 6
    bms[4] = 6;				// section 6
    bms[5] = 0;				// has bitmap

    cbits = bms + 6;
    c = start = 0;
    imask = 128;
    for (i = 0; i < nn; i++) {
	if (data[i] < UNDEFINED_LOW || data[i] > UNDEFINED_HIGH) {
	    c += imask;
	    data[start++] = data[i];
	}
	if ((imask >>= 1) == 0) {
	    *cbits++ = c;
	    c = 0;
	    imask = 128;
	}
    }
    if (imask != 128) *cbits = c;
    *ndata = start;
    if (nn == start) {
	free (bms);
	bms = (unsigned char *) malloc(6);
	uint_char(6, bms);		// length of section 6
	bms[4] = 6;			// section 5
	bms[5] = 255;			// no bitmap
    }
    return bms;
}


/*
 * HEADER:100:set_bin_prec:misc:1:X set number of binary bits for grib_out packing
 */

int f_set_bin_prec(ARG1) {
    int i;

    struct local_struct {
        int nbits;
    };
    struct local_struct *save;

    if (mode == -1) {

        i = atoi(arg1);
	if (i < 0) i = 0;
	if (i > max_bits) i = max_bits;

        save = (struct local_struct *)malloc( sizeof(struct local_struct));
        if (save == NULL) fatal_error("set_bit_prec  memory allocation ","");
        *local = save;
        save->nbits = wanted_bits = i;
	use_scale = 0;
    }
    if (mode >= 0) {
	save = *local;
	wanted_bits = save->nbits;
	use_scale = 0;
    }
    return 0;
}

/*
 * HEADER:100:set_scaling:misc:2:set decimal scaling=X binary scaling=Y for grib_out packing
 */

int f_set_scaling(ARG2) {

    struct local_struct {
        int dec, bin;
    };
    struct local_struct *save;

    if (mode == -1) {
        save = (struct local_struct *)malloc( sizeof(struct local_struct));
        if (save == NULL) fatal_error("set_bit_prec  memory allocation ","");
        *local = save;

        dec_scale = save->dec = atoi(arg1);
        bin_scale = save->bin = atoi(arg2);

        use_scale = 1;
    }
    if (mode >= 0) {
        save = *local;
        dec_scale = save->dec;
        bin_scale = save->bin;
        use_scale = 1;
    }
    return 0;
}



/*
 * make sec 5 and 7 using simple packing
 */


int mk_sec5_7(float *data, unsigned int n, unsigned char **sec5, unsigned char **sec7) {

    float min_val, max_val, ref, ncep_min_val;
    int nbits, binary_scale, j;
    unsigned int i;
    double frange, scale, dec_factor;
    unsigned long int sec5_size, sec7_size;
    unsigned char *p;

    for (max_val = min_val = data[0], i = 1; i < n; i++) {
	if (min_val > data[i]) min_val = data[i];
        if (max_val < data[i]) max_val = data[i];
    }
    binary_scale = bin_scale;
    ncep_min_val = min_val;

    if (use_scale == 0) {
        /* ecmwf style */
        ref = min_val;
        frange = max_val - ref;
        dec_scale = 0;
        if (frange != 0.0) {
            frexp(frange, &j);
            binary_scale = j - wanted_bits;
            nbits = wanted_bits;
            scale = ldexp(1.0, -binary_scale);
            frange = floor((max_val-ref)*scale + 0.5);
            frexp(frange, &j);
            if (j != nbits) binary_scale++;
        }
        else {
            binary_scale = nbits = 0;
            scale = 1;
        }
    }
    else {
        if (dec_scale) {
            dec_factor = Int_Power(10.0, -dec_scale);
            min_val *= dec_factor;
            max_val *= dec_factor;
            for (i = 0; i < n; i++) {
                data[i] *= dec_factor;
            }
        }

        scale = ldexp(1.0, -binary_scale);
        ref = floor(min_val*scale)/scale;
        j = (int) (max_val - ref)*scale + 0.5;
        frange = (double) j;
        frexp(frange, &nbits);
        if (nbits > max_bits) {
            binary_scale += (nbits - max_bits);
            nbits = max_bits;
        }
    }
    if (binary_scale) {
        scale = ldexp(1.0, -binary_scale);
        for (i = 0; i < n; i++) {
            data[i] = (data[i] - ref)*scale;
        }
    }
    else {
        for (i = 0; i < n; i++) {
            data[i] = data[i] - ref;
        }
    }

    sec5_size = 21;
    sec7_size = 5 + (nbits * (n / 8)) + (nbits * (n % 8) + 7) / 8;

    // section 7
    *sec7 = p = (unsigned char *) malloc(sec7_size);
    uint_char(sec7_size, p);
    p[4] = 7;
    flist2bitstream(data,p + 5,n,nbits);

    // section 5

    // fix for buggy NCEP decoders
    // for constant fields, they ignore the decimal scaling
    if (nbits == 0) {
	dec_scale = binary_scale = 0;
	ref = ncep_min_val;
    }

    *sec5 = p = (unsigned char *) malloc(sec5_size);
    uint_char(sec5_size, p);		// length of section 5
    p[4] = 5;				// section 5
    uint_char(n, p+5);			// number of defined points
    uint2_char(0,p+9);			// template 5.0
    flt2ieee(ref,p+11);			// ieee reference value
    int2_char(binary_scale,p+15);
    int2_char(-dec_scale,p+17);
    p[19] = nbits;
    p[20] = 0;				// template 5.1 - set to floating

    return 0;
}



/*
 * grib: convert linear list of ints to a bitstream
 *
 * in public domain : 2007 Wesley Ebisuzaki
 */


static unsigned int mask[] = {0,1,3,7,15,31,63,127,255};

void flist2bitstream(float *list, unsigned char *bitstream, unsigned int ndata, int nbits) 
{

    int cbits, jbits;
    unsigned int j, c;

    if (nbits == 0) {
	return;
    }
    if (nbits < 0) {
	fprintf(stderr,"nbits < 0!  nbits = %d\n", nbits);
	exit(0);
    }

    cbits = 8;
    c = 0;
    while (ndata-- > 0) {
	/* note float -> unsigned int .. truncate */
        j = (unsigned int) (*list++ + 0.5);
	jbits = nbits;
	while (cbits <= jbits) {
	    if (cbits == 8) {
	        jbits -= 8;
	        *bitstream++ = (j >> jbits) & 255;
	    }
	    else {
	        jbits -= cbits;
	        *bitstream++ = (c << cbits) + ((j >> jbits) & mask[cbits]);
		cbits = 8;
	        c = 0;
	    }
	}
	/* now jbits < cbits */
	if (jbits) {
	    c = (c << jbits) + (j & mask[jbits]);
	    cbits -= jbits;
	}
    }
    if (cbits != 8) *bitstream++ = c << cbits;
}

int set_order(unsigned char **sec, enum output_order_type order) {
    int flag_3_4;

    if (order == raw) return 0;
    flag_3_4 = flag_table_3_4(sec) & 15;
    if (order == wesn) return set_flag_table_3_4(sec, flag_3_4 | (4 << 4));
    if (order == wens) return set_flag_table_3_4(sec, flag_3_4);
    return 1;
}

/*
 * write sections as a grib message
 */

int wrt_sec(unsigned char *sec0, unsigned char *sec1, unsigned char *sec2, 
   unsigned char *sec3, unsigned char *sec4, unsigned char *sec5, 
   unsigned char *sec6, unsigned char *sec7, FILE *out) {

    unsigned long int size;
    unsigned char s[8];
    int i;

    size = (unsigned long int) GB2_Sec0_size + GB2_Sec8_size +
         (sec1 ? uint4(sec1) : 0) +
         (sec2 ? uint4(sec2) : 0) +
         (sec3 ? uint4(sec3) : 0) +
         (sec4 ? uint4(sec4) : 0) +
         (sec5 ? uint4(sec5) : 0) +
         (sec6 ? uint4(sec6) : 0) +
         (sec7 ? uint4(sec7) : 0);

    fwrite((void *) sec0, sizeof(char), 8, out);
    uint8_char(size, s);
    fwrite((void *) s, sizeof(char), 8, out);

    if (sec1) {
	i = uint4(sec1);
	if (fwrite((void *)sec1, sizeof(char), i, out) != i) return 1;
    }
    if (sec2) {
	i = uint4(sec2);
	if (fwrite((void *)sec2, sizeof(char), i, out) != i) return 1;
    }
    if (sec3) {
	i = uint4(sec3);
	if (fwrite((void *)sec3, sizeof(char), i, out) != i) return 1;
    }
    if (sec4) {
	i = uint4(sec4);
	if (fwrite((void *)sec4, sizeof(char), i, out) != i) return 1;
    }
    if (sec5) {
	i = uint4(sec5);
	if (fwrite((void *)sec5, sizeof(char), i, out) != i) return 1;
    }
    if (sec6) {
	i = uint4(sec6);
	if (fwrite((void *)sec6, sizeof(char), i, out) != i) return 1;
    }
    if (sec7) {
	i = uint4(sec7);
	if (fwrite((void *)sec7, sizeof(char), i, out) != i) return 1;
    }
    s[0] = s[1] = s[2] = s[3] = 55; /* s = "7777" */
    if (fwrite((void *) s, sizeof(char), 4, out) != 4) return 1;

    return 0;
}
